/*
 * multilevel.c
 */

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <GL/glaux.h>
#else
#include <gtk/gtk.h>
#include <gtkgl/gtkglarea.h>
#include <gdk/gdkkeysyms.h>
#endif

#include <GL/gl.h>
#include <GL/glu.h>

#include <math.h>


#include "multilevel.h"
#include "mc.h"
#include "mesh.h"
#include "voxels.h"
#include "trackball.h"


/******************************************************************/

#define VIEW_ASPECT 1.0


int beginx, beginy;
float quat[4];
float zoom = 45;
int detail = 1;
int grid_depth = 0;
unsigned int submeshes = 0xFFFFFFFF;
int shading = 0;
float expand = 0.0;
int active = 0;
float active_x, active_y, active_z;


int
init_gl()
{
  GLfloat light0_pos[4]   = { -50.0, 50.0, 0.0, 0.0 };
  GLfloat light0_color[4] = { 1.6, 1.6, 1.6, 1.0 }; /* white light */
  GLfloat light1_pos[4]   = {  50.0, 50.0, 0.0, 0.0 };
  GLfloat light1_color[4] = { .8, .8, 2.0, 1.0 };  /* cold blue light */
  GLfloat lighta_color[4] = { .3, .3, .3, 1.0 }; /* ambient light */

  glEnable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);

  if (!shading)
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

  glEnable(GL_DITHER);
  glShadeModel(GL_SMOOTH);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
  glHint(GL_POLYGON_SMOOTH_HINT, GL_FASTEST);

  glLightfv(GL_LIGHT0, GL_POSITION, light0_pos);
  glLightfv(GL_LIGHT0, GL_DIFFUSE,  light0_color);  
  glLightfv(GL_LIGHT1, GL_POSITION, light1_pos);
  glLightfv(GL_LIGHT1, GL_DIFFUSE,  light1_color);
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHT1);
  glEnable(GL_LIGHTING);

  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lighta_color);
    
  glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
  glEnable(GL_COLOR_MATERIAL);

  trackball(quat, 0, 0, 0, 0);

  return TRUE;
}


void
generate_mesh()
{
  int x, y, z;

  /* March cubes. */
  for (x = 0; x < DIMX; x += GRANULARITY) {
    for (y = 0; y < DIMY; y += GRANULARITY) {
      for (z = 0; z < DIMZ; z += GRANULARITY) {
        adaptive_march_cube(0, x, y, z, GRANULARITY, GRANULARITY, GRANULARITY);
#if 0
        adaptive_march_cube(max_depth_at_point(x+GRANULARITY/2,y+GRANULARITY/2,z+GRANULARITY/2),
                            x, y, z, GRANULARITY, GRANULARITY, GRANULARITY);
#endif
      }
    }
  }
  /* The adaptive algorithm leaves gaps --- patch them up. */
  adaptive_patch_mesh();
  g_print("generated %i tris\n", triangles_length);

#if 0
  recompute_vertex_normals();
#endif
}

void
draw_bounding_cube_vertex(float x, float y, float z)
{
#ifndef WIN32
  glColor3f(x, y, z);
#endif
  glVertex3f(x, y, z);
}

void
draw_bounding_cube()
{
  float prev_ambient[4], cube_ambient[4] = { 1.0, 1.0, 1.0, 1.0 };

  glGetFloatv(GL_LIGHT_MODEL_AMBIENT, prev_ambient);
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, cube_ambient);

  glColor3f(0, 0, 0);
  glBegin(GL_LINES);
  draw_bounding_cube_vertex(0,0,0);
  draw_bounding_cube_vertex(0,1,0);
  draw_bounding_cube_vertex(0,1,0);
  draw_bounding_cube_vertex(1,1,0);
  draw_bounding_cube_vertex(1,1,0);
  draw_bounding_cube_vertex(1,0,0);
  draw_bounding_cube_vertex(1,0,0);
  draw_bounding_cube_vertex(0,0,0);

  draw_bounding_cube_vertex(0,0,0);
  draw_bounding_cube_vertex(0,0,1);
  draw_bounding_cube_vertex(0,1,0);
  draw_bounding_cube_vertex(0,1,1);
  draw_bounding_cube_vertex(1,1,0);
  draw_bounding_cube_vertex(1,1,1);
  draw_bounding_cube_vertex(1,0,0);
  draw_bounding_cube_vertex(1,0,1);

  draw_bounding_cube_vertex(0,0,1);
  draw_bounding_cube_vertex(0,1,1);
  draw_bounding_cube_vertex(0,1,1);
  draw_bounding_cube_vertex(1,1,1);
  draw_bounding_cube_vertex(1,1,1);
  draw_bounding_cube_vertex(1,0,1);
  draw_bounding_cube_vertex(1,0,1);
  draw_bounding_cube_vertex(0,0,1);
  glEnd();

  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, prev_ambient);
}

void
draw_grid()
{
  int n, i, j;
  n = 1 << grid_depth;

  glColor3f(0, 0, 0);
  glBegin(GL_LINES);

  for (i = 1; i < n; i++) {
    for (j = 1; j < n; j++) {
      glVertex3f(0, (float)i/n, (float)j/n);
      glVertex3f(1, (float)i/n, (float)j/n);
    }
  }
  for (i = 1; i < n; i++) {
    for (j = 1; j < n; j++) {
      glVertex3f((float)i/n, 0, (float)j/n);
      glVertex3f((float)i/n, 1, (float)j/n);
    }
  }
  for (i = 1; i < n; i++) {
    for (j = 1; j < n; j++) {
      glVertex3f((float)i/n, (float)j/n, 0);
      glVertex3f((float)i/n, (float)j/n, 1);
    }
  }

  glEnd();
}

void
draw_cubes()
{
  int i;
  static int inited = 0;

  /* Init volume data if necessary. */
  if (!inited) {
    init_voxels();
    generate_mesh();
    inited = 1;
  }

  /* Draw bounding cube. */
  draw_bounding_cube();

  /* Draw grid if on. */
  if (grid_depth > 0)
    draw_grid();

  /* Draw the mesh. */
#ifdef WIN32
  /* Something about the Win32 GL is buggy, and doesn't recognize glColor in between
   * glBegin and glEnd.  Thus, I can't have per-vertex color for debugging. */
  glColor3f(1.6, 1.9, 1.9);
#endif

  glBegin(GL_TRIANGLES);
  for (i = 0; i < triangles_length; i++) {
    if (!(submeshes & triangles[i].submeshes)) continue;
#ifndef WIN32
    glColor3f (triangles[i].v[2]->cr,
               triangles[i].v[2]->cg,
               triangles[i].v[2]->cb);
#endif
    glNormal3f(triangles[i].v[2]->nx,
               triangles[i].v[2]->ny,
               triangles[i].v[2]->nz);
    glVertex3f(triangles[i].v[2]->x + expand*triangles[i].v[2]->nx,
               triangles[i].v[2]->y + expand*triangles[i].v[2]->ny,
               triangles[i].v[2]->z + expand*triangles[i].v[2]->nz);
#ifndef WIN32
    glColor3f (triangles[i].v[1]->cr,
               triangles[i].v[1]->cg,
               triangles[i].v[1]->cb);
#endif
    glNormal3f(triangles[i].v[1]->nx,
               triangles[i].v[1]->ny,
               triangles[i].v[1]->nz);
    glVertex3f(triangles[i].v[1]->x + expand*triangles[i].v[1]->nx,
               triangles[i].v[1]->y + expand*triangles[i].v[1]->ny,
               triangles[i].v[1]->z + expand*triangles[i].v[1]->nz);
#ifndef WIN32
    glColor3f (triangles[i].v[0]->cr,
               triangles[i].v[0]->cg,
               triangles[i].v[0]->cb);
#endif
    glNormal3f(triangles[i].v[0]->nx,
               triangles[i].v[0]->ny,
               triangles[i].v[0]->nz); 
    glVertex3f(triangles[i].v[0]->x + expand*triangles[i].v[0]->nx,
               triangles[i].v[0]->y + expand*triangles[i].v[0]->ny,
               triangles[i].v[0]->z + expand*triangles[i].v[0]->nz);
  }
  glEnd();
}


/********************************************************************************/
/* Code to calculate the "area of interest" as defined by the pointer location. */
int
pointer_to_surface(float px, float py,
                   float *x, float *y, float *z)
{
  GLdouble winx, winy, winz, objx, objy, objz, objx2, objy2, objz2;
  GLdouble model_matrix[16], proj_matrix[16];
  GLint viewport[4];
  int i, j, k;
  GLdouble wxmin, wxmax, wymin, wymax, wzmin, wzmax;
  GLdouble zstep;
  float v[3];

  glGetDoublev(GL_MODELVIEW_MATRIX, model_matrix);
  glGetDoublev(GL_PROJECTION_MATRIX, proj_matrix);
  glGetIntegerv(GL_VIEWPORT, viewport);

  /* Find the min and max window coordinates. */
  wxmin = wymin = wzmin =  3.402823466e+38;
  wxmax = wymax = wzmax = -3.402823466e+38;
  for (i = 0; i < 2; i++) {
    for (j = 0; j < 2; j++) {
      for (k = 0; k < 2; k++) {
        gluProject(i, j, k, model_matrix, proj_matrix, viewport, &winx, &winy, &winz);
        if (winx < wxmin) wxmin = winx;
        if (winy < wymin) wymin = winy;
        if (winz < wzmin) wzmin = winz;
        if (winx > wxmax) wxmax = winx;
        if (winy > wymax) wymax = winy;
        if (winz > wzmax) wzmax = winz;
      }
    }
  }

  /* Check that the pointer is within the bounding box at all. */
  winx = px;
  winy = py;
  if (winx < wxmin || winx > wxmax || winy < wymin || winy > wymax)
    return 0;

  /* Calibrate our step size to the window coordinates.  Note sure if this could
   * be done statically, but it doesn't really matter. */
  gluUnProject(winx, winy, wzmin, model_matrix, proj_matrix, viewport, &objx , &objy , &objz);
  gluUnProject(winx, winy, wzmax, model_matrix, proj_matrix, viewport, &objx2, &objy2, &objz2);
  zstep = (wzmax - wzmin)
    / (2 * MIN(DIMX, MIN(DIMY, DIMZ))
       * sqrt((objx2-objx)*(objx2-objx) + (objy2-objy)*(objy2-objy) + (objz2-objz)*(objz2-objz)));

  /* Walk through the voxel space till we hit the surface. */
  for (winz = wzmin; winz <= wzmax; winz += zstep) {
    gluUnProject(winx, winy, winz, model_matrix, proj_matrix, viewport, &objx , &objy , &objz);

    v[0] = objx;
    v[1] = objy;
    v[2] = objz;
    if (   v[0] >= 0.0 && v[0] < 1.0
        && v[1] >= 0.0 && v[1] < 1.0
        && v[2] >= 0.0 && v[2] < 1.0
        && voxel_eval(v) > 127) {
      *x = objx;
      *y = objy;
      *z = objz;
      return 1;
    }
  }

  /* We never hit the surface. */
  return 0;
}

/* Code to calculate the max depth to recurse to at a point. */
#define MINDEPTH 1
#define MAXDEPTH 8
#define WIDTH  (GRANULARITY/2)

int
max_depth_at_point(int x, int y, int z)
{
  float r;
  if (active) {
    r = sqrt(  (x - active_x*(DIMX+1))*(x - active_x*(DIMX+1))
             + (y - active_y*(DIMY+1))*(y - active_y*(DIMY+1))
             + (z - active_z*(DIMZ+1))*(z - active_z*(DIMZ+1)))
      / WIDTH;
    return r < 1.0 ? MAXDEPTH : (float)MAXDEPTH/(r*r) + MINDEPTH;
  } else {
    return MINDEPTH;
  }
}

int
max_depth_in_box(int  x, int  y, int  z,
                 int sx, int sy, int sz)
{
  int max = MINDEPTH, val;
  int i, j, k;

  if (active &&
      x <= active_x*(DIMX+1) && active_x*(DIMX+1) <= x+sx &&
      y <= active_y*(DIMY+1) && active_y*(DIMY+1) <= y+sy &&
      z <= active_z*(DIMZ+1) && active_z*(DIMZ+1) <= z+sz) {
    return MAXDEPTH;
  }

  for (i = 0; i < 2; i++) {
    for (j = 0; j < 2; j++) {
      for (k = 0; k < 2; k++) {
        val = max_depth_at_point(x+i*sx, y+j*sy, z+k*sz);
        if (val > max) max = val;
      }
    }
  }
  return max;
}


/******************************************************************/
/* Platform-independent event handling stuff. */

void
draw_gl()
{
  GLfloat m[4][4];

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(zoom, VIEW_ASPECT, 1, 100);
  glMatrixMode(GL_MODELVIEW);

  glClearColor(0.3, 0.4, 0.6, 1);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glLoadIdentity();
  glTranslatef(0, 0, -30);
  glScalef(10, 10, 10);
  build_rotmatrix(m, quat);
  glMultMatrixf(&m[0][0]);
  glTranslatef(-0.5, -0.5, -0.5);

  draw_cubes();
}

void
reshape_gl(int width,
           int height)
{
  glViewport(0, 0, width, height);
}

/* keep track of pointer update */
void
pointer_update(int x,
               int y,
               int area_width,
               int area_height)
{
  active = pointer_to_surface(x, area_height-y, &active_x, &active_y, &active_z);
  if (active) {
    clear_triangles();
    clear_vertices();
    generate_mesh();
  }
}

/* virtual trackball */
void
button_press(int x,
             int y)
{
  /* beginning of drag, reset mouse position */
  beginx = x;
  beginy = y;
}

void
drag_spin(int x,
          int y,
          int area_width,
          int area_height)
{
  /* drag in progress, simulate trackball */
  float spin_quat[4];
  trackball(spin_quat,
            (2.0 *     beginx -       area_width) / area_width,
            (     area_height - 2.0 *     beginy) / area_height,
            (           2.0*x -       area_width) / area_width,
            (     area_height -            2.0*y) / area_height);
  add_quats(spin_quat, quat, quat);

  beginx = x;
  beginy = y;
}

void
drag_zoom(int x,
          int y,
          int area_width,
          int area_height)
{
  /* zooming drag */
  zoom += ((float)(y - beginy) / area_height) * 40;
  if (zoom < 5) zoom = 5;
  if (zoom > 120) zoom = 120;

  beginx = x;
  beginy = y;
}

int
drag_detail(int x,
            int y,
            int area_width,
            int area_height)
{
  /* detail drag */
  static float fdetail = 1.0;
  fdetail += ((float)(beginy - y) / area_height) * 20;
  if (fdetail < 1.0) fdetail = 1.0;
  if (fdetail > GRANULARITY) fdetail = GRANULARITY;

  beginx = x;
  beginy = y;

  if (detail != (int)fdetail) {
    detail = (int)fdetail;
    g_print("detail %i\n", detail);

    /* Regenerate volume data with correct ramp. */
    init_voxels();

    return TRUE;
  }

  return FALSE;
}

int
toggle_submesh(int submesh)
{
  g_print("toggle_submesh %i\n", submesh);
  if (submesh == 0) {
    submeshes ^= 1 << 9;
  } else {
    submeshes ^= 1 << (submesh-1);
  }
  return 0;
}

int
toggle_shading()
{
  g_print("toggle_shading\n");
  shading = !shading;
  if (shading) {
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  } else {
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
  }
  return 0;
}

int
increment_grid_depth(int increment_by)
{
  grid_depth += increment_by;
  g_print("increment_grid_depth %i to %i\n", increment_by, grid_depth);
  return 0;
}

int
expand_mesh(float expand_by)
{
  expand += expand_by;
  g_print("expand_mesh %f to %f\n", expand_by, expand);
  return 0;
}



/*****************************************************************/
/* GTK bindings and code */
#ifndef WIN32
gint
glarea_expose(GtkWidget      *widget,
              GdkEventExpose *event)
{
  GtkGLArea *glarea = GTK_GL_AREA(widget);
  static int inited = 0;

  if (inited == 0) {
    if (gtk_gl_area_make_current(glarea) == FALSE) {
      g_print("Couldn't make GL area current for initialization\n");
      return FALSE;
    }

    if (!init_gl())
      return TRUE;
    inited = 1;
  }

  if (event->count > 0)
    return TRUE;

  if (gtk_gl_area_make_current(glarea) == FALSE)
    return TRUE;

  draw_gl();

  gtk_gl_area_swapbuffers(glarea);

  return TRUE;
}

gint
glarea_configure(GtkWidget         *widget,
                 GdkEventConfigure *event)
{
  if (gtk_gl_area_make_current(GTK_GL_AREA(widget))) {
    reshape_gl(widget->allocation.width, widget->allocation.height);
  }
  return TRUE;
}


/* virtual trackball */
gint
glarea_button_press(GtkWidget      *widget,
                    GdkEventButton *event)
{
  if (event->button == 1 || event->button == 2 || event->button == 3) {
    button_press(event->x, event->y);
    return TRUE;
  }
  return FALSE;
}

gint
glarea_motion_notify(GtkWidget      *widget,
                     GdkEventMotion *event)
{
  int x, y;
  GdkRectangle area;
  GdkModifierType state;

  if (event->is_hint) {
    gdk_window_get_pointer(event->window, &x, &y, &state);
  } else {
    x = event->x;
    y = event->y;
    state = event->state;
  }
  
  area.x = 0;
  area.y = 0;
  area.width  = widget->allocation.width;
  area.height = widget->allocation.height;

  if (state & GDK_BUTTON1_MASK) {
    drag_spin(x, y, area.width, area.height);
    /* orientation has changed, redraw mesh */
    gtk_widget_draw(widget, &area);
  } else if (state & GDK_BUTTON2_MASK) {
    drag_zoom(x, y, area.width, area.height);
    /* zoom has changed, redraw mesh */
    gtk_widget_draw(widget, &area);
  } else if (state & GDK_BUTTON3_MASK) {
    if (drag_detail(x, y, area.width, area.height)) {
      /* detail has changed, redraw mesh */
      gtk_widget_draw(widget, &area);
    }
  } else {
    /* set active area */
    pointer_update(x, y, area.width, area.height);
    gtk_widget_draw(widget, &area);
  }

  return TRUE;
}

gint
glarea_key_press(GtkWidget   *widget,
                 GdkEventKey *event,
                 gpointer     data)
{
  if (GDK_0 <= event->keyval && event->keyval <= GDK_9) {
    toggle_submesh(event->keyval - GDK_0);
    gtk_widget_draw(widget, NULL);
  } else if (event->keyval == GDK_s) {
    toggle_shading();
    gtk_widget_draw(widget, NULL);
  } else if (event->keyval == GDK_e) {
    expand_mesh(0.01);
    gtk_widget_draw(widget, NULL);
  } else if (event->keyval == GDK_r) {
    expand_mesh(-0.01);
    gtk_widget_draw(widget, NULL);
  } else if (event->keyval == GDK_i) {
    increment_grid_depth(1);
    gtk_widget_draw(widget, NULL);
  } else if (event->keyval == GDK_o) {
    increment_grid_depth(-1);
    gtk_widget_draw(widget, NULL);
  }
  return TRUE;
}


/* main function, only used in gtk version */
GtkWidget *window, *frame;
GtkGLArea *glarea;

int
main(int   argc,
     char *argv[])
{
  gtk_init(&argc, &argv);

  if (gdk_gl_query() == FALSE) {
    g_print("OpenGL not supported\n");
    return 1;
  }

  frame = gtk_aspect_frame_new(NULL, 0.5, 0.5, VIEW_ASPECT, FALSE);

  glarea = (GtkGLArea*)
    gtk_gl_area_new_vargs(NULL,
                          GDK_GL_RGBA,
                          GDK_GL_DOUBLEBUFFER,
                          GDK_GL_DEPTH_SIZE, 1,
                          GDK_GL_NONE);
  if (glarea == NULL) {
    g_print("Can't create GtkGLArea widget\n");
    return 1;
  }
  gtk_widget_set_events(GTK_WIDGET(glarea),
                        GDK_EXPOSURE_MASK |
                        GDK_BUTTON_PRESS_MASK |
                        GDK_BUTTON_RELEASE_MASK |
                        GDK_POINTER_MOTION_MASK |
                        GDK_POINTER_MOTION_HINT_MASK |
                        GDK_KEY_PRESS_MASK);
  gtk_signal_connect(GTK_OBJECT(glarea), "expose_event",
                     GTK_SIGNAL_FUNC(glarea_expose), NULL);
  gtk_signal_connect(GTK_OBJECT(glarea), "configure_event",
                     GTK_SIGNAL_FUNC(glarea_configure), NULL);
  gtk_signal_connect(GTK_OBJECT(glarea), "button_press_event",
                     GTK_SIGNAL_FUNC(glarea_button_press), NULL);
  gtk_signal_connect(GTK_OBJECT(glarea), "motion_notify_event",
                     GTK_SIGNAL_FUNC(glarea_motion_notify), NULL);

  gtk_widget_set_usize(GTK_WIDGET(glarea), 600, 600 / VIEW_ASPECT);

  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(window), "Multilevel Marching Cubes");
  gtk_signal_connect(GTK_OBJECT(window), "destroy",
                     GTK_SIGNAL_FUNC(gtk_main_quit), NULL);
  gtk_signal_connect(GTK_OBJECT(window), "key_press_event",
                     GTK_SIGNAL_FUNC(glarea_key_press), NULL);

  gtk_container_add(GTK_CONTAINER(window), frame);
  gtk_container_add(GTK_CONTAINER(frame), GTK_WIDGET(glarea));
  gtk_widget_show(GTK_WIDGET(glarea));
  gtk_widget_show(frame);
  gtk_widget_show(window);

  gtk_main();

  return 0;
}
#endif


/********************************************************************/
/* Win32 code */
#ifdef WIN32

void CALLBACK
aux_expose(/* int junk1, int junk2 */)
{
  static int inited = 0;

  if (inited == 0) {
    if (!init_gl())
      return;
    inited = 1;
  }

  draw_gl();

  auxSwapBuffers();
}

int aux_area_width = 600, aux_area_height = 600;

void CALLBACK
aux_reshape(int width,
            int height)
{
  aux_area_width = width;
  aux_area_height = height;
  reshape_gl(width, height);
}

void CALLBACK
aux_mouse_press(AUX_EVENTREC *event)
{
  button_press(event->data[AUX_MOUSEX], event->data[AUX_MOUSEY]);
}

void CALLBACK
aux_mouse_motion0(AUX_EVENTREC *event)
{
  if (event->data[AUX_MOUSESTATUS] == 0) {
    pointer_update(event->data[AUX_MOUSEX], event->data[AUX_MOUSEY],
                   aux_area_width, aux_area_height);
    aux_expose();
  }
}

void CALLBACK
aux_mouse_motion1(AUX_EVENTREC *event)
{
  drag_spin(event->data[AUX_MOUSEX], event->data[AUX_MOUSEY],
            aux_area_width, aux_area_height);
  aux_expose();
}

void CALLBACK
aux_mouse_motion2(AUX_EVENTREC *event)
{
  drag_zoom(event->data[AUX_MOUSEX], event->data[AUX_MOUSEY],
            aux_area_width, aux_area_height);
  aux_expose();
}

void CALLBACK
aux_mouse_motion3(AUX_EVENTREC *event)
{
  if (drag_detail(event->data[AUX_MOUSEX], event->data[AUX_MOUSEY],
                  aux_area_width, aux_area_height))
    aux_expose();
}


#define GENERATE_AUX_KEY_PRESS_FUNC(i) \
  void CALLBACK \
  aux_key_press_##i(void) \
  { \
    toggle_submesh(i); \
    aux_expose(); \
  }
GENERATE_AUX_KEY_PRESS_FUNC(0)
GENERATE_AUX_KEY_PRESS_FUNC(1)
GENERATE_AUX_KEY_PRESS_FUNC(2)
GENERATE_AUX_KEY_PRESS_FUNC(3)
GENERATE_AUX_KEY_PRESS_FUNC(4)
GENERATE_AUX_KEY_PRESS_FUNC(5)
GENERATE_AUX_KEY_PRESS_FUNC(6)
GENERATE_AUX_KEY_PRESS_FUNC(7)
GENERATE_AUX_KEY_PRESS_FUNC(8)
GENERATE_AUX_KEY_PRESS_FUNC(9)

void CALLBACK
aux_key_press_s(void)
{
  toggle_shading();
  aux_expose();
}




int
main(int argc,
     char *argv[])
{
  auxInitDisplayMode(AUX_RGB | AUX_DOUBLE);
  auxInitPosition(0, 0, aux_area_width, aux_area_height);
  auxInitWindow("foo");

  auxReshapeFunc((AUXRESHAPEPROC) aux_reshape);

  auxMouseFunc(AUX_LEFTBUTTON,   AUX_MOUSEDOWN, (AUXMOUSEPROC) aux_mouse_press);
  auxMouseFunc(AUX_MIDDLEBUTTON, AUX_MOUSEDOWN, (AUXMOUSEPROC) aux_mouse_press);
  auxMouseFunc(AUX_RIGHTBUTTON,  AUX_MOUSEDOWN, (AUXMOUSEPROC) aux_mouse_press);
  auxMouseFunc(0,                AUX_MOUSELOC,  (AUXMOUSEPROC) aux_mouse_motion0);
  auxMouseFunc(AUX_LEFTBUTTON,   AUX_MOUSELOC,  (AUXMOUSEPROC) aux_mouse_motion1);
  auxMouseFunc(AUX_MIDDLEBUTTON, AUX_MOUSELOC,  (AUXMOUSEPROC) aux_mouse_motion2);
  auxMouseFunc(AUX_RIGHTBUTTON,  AUX_MOUSELOC,  (AUXMOUSEPROC) aux_mouse_motion3);

  auxKeyFunc('0', (AUXKEYPROC) aux_key_press_0);
  auxKeyFunc('1', (AUXKEYPROC) aux_key_press_1);
  auxKeyFunc('2', (AUXKEYPROC) aux_key_press_2);
  auxKeyFunc('3', (AUXKEYPROC) aux_key_press_3);
  auxKeyFunc('4', (AUXKEYPROC) aux_key_press_4);
  auxKeyFunc('5', (AUXKEYPROC) aux_key_press_5);
  auxKeyFunc('6', (AUXKEYPROC) aux_key_press_6);
  auxKeyFunc('7', (AUXKEYPROC) aux_key_press_7);
  auxKeyFunc('8', (AUXKEYPROC) aux_key_press_8);
  auxKeyFunc('9', (AUXKEYPROC) aux_key_press_9);
  auxKeyFunc(AUX_s, (AUXKEYPROC) aux_key_press_s);

  auxExposeFunc((AUXEXPOSEPROC) aux_expose);

  auxMainLoop((AUXMAINPROC) aux_expose);
  return 0;
}
#endif
